unit ULinkage;



interface

Uses Sysutils, Classes, Controls, ExtCtrls, Types, Graphics, UGeometry,
        U_ConnectDlg;

type

  TMovementType=(Pinned, Float, Trace); {Ways that a connection can move}

  TConnection=class(TObject)
    ID: string;
    Loc:TRealPoint;
    
    movementType:TMovementType;
    ConnectColor:TColor;

    moved:boolean;  {flag indicating that this connection point has been moved
                     in this move iteration}
    tracepoints:array of tpoint;
    pt,prevpt:TPoint; {used for trace position checking}
    tracecomplete:boolean;{true after a complete trace cycle has been completed}

    procedure resettrace;  {reset any trace information for this connection}
    procedure setloc(newloc:Trealpoint);
    procedure save(fstream:TFilestream);
    procedure Load(fstream:TFilestream);
  end;

  TForcetype=(Linear, Rotational,None);

  TForce=class(TObject)
    FType:TForceType;
    Strength:integer;
    dir:extended;  {angle}
  end;

  TRod= class(TObject)
    id:string;
    p1,p2:TRealPoint;
    theta:extended;
    Len:extended;
    rodcolor:TColor;
    Connections:array of String[3];
    Force:TForce;
    w:integer;
    destructor destroy; override;
    procedure save(fstream:TFileStream);
    procedure load(fstream:TFilestream);
  end;

  TConnections=class(TStringlist)
    constructor create;
    procedure save(fstream:TFileStream);
    procedure load(fstream:TFilestream);
  end;

  TDragmode=(operate, modify, nodrag);

  TDragrec=record
    nbr,pnbr:integer;
  end;

  TLinkage=class(TPaintbox)
  Public
    rods:array of TRod;
    selectednbr:integer;
    dragmode:Tdragmode;
    dragged:TConnection;
    draggedrods:array of TDragrec;
    connectList:TConnections;
    constructor create(Aowner:TComponent; Proto:TPaintbox);    
    destructor destroy; override;
    function getconnection(s:string):Tconnection;
    procedure MoveTo(connect:TConnection; newloc:TRealPoint);
    function FindFixedConnection(Rod:Trod):integer;
    function makeconnection(newId:string;newloc:TRealPoint;
                       newmovement:TMovementtype; newcolor:TColor):integer;
    procedure LinkMouseDown(Sender: TObject; Button: TMouseButton;
                       Shift: TShiftState; X, Y: Integer);
    procedure LinkMouseMove(Sender: TObject; Shift: TShiftState; X,Y: Integer);
    procedure LinkMouseUp(Sender: TObject; Button: TMouseButton;
                      Shift: TShiftState; X, Y: Integer);
    procedure Paint(Sender: TObject); 
    procedure DrawLinks;
    procedure Save(fname:string);
    procedure Load(fname:string);

  end;

implementation

procedure writestr(const fstream:TFilestream; const s:string);
var
  i:integer;
begin
  with fstream do
  begin
    i:=length(s);
    writebuffer(i,sizeof(i));
    if i>0 then writebuffer(s[1],i);
  end;
end;

function readstr(const fstream:TFilestream):string;
var
  i:integer;
  s:string;
begin
  with fstream do
  begin
    readbuffer(i,sizeof(i));
    setlength(s,i);
    if i>0 then readbuffer(s[1],i);
    result:=s;
  end;
end;

{************* TConnection.moveto ************}
procedure TLinkage.Moveto(connect:TConnection; newloc:TRealpoint);
var
  oldloc:Trealpoint;
  i:integer;
  connect1:TConnection;
  L:TRealLine;
  len:extended;
  p:TPoint;
begin
  with connect do
  begin
    oldloc:=loc;
    loc:=newloc;
    {we have moved the connection, now run through the rods and
     set p1 to connections[0] and p2 to connections[high(connections)]
    }

    for i:=0 to high(draggedrods) do
    with draggedrods[i] do
    with rods[nbr] do
    begin
      if (pnbr=1) or (pnbr=2) then
      begin
        Len:=dist(p1,p2);
        theta:=gettheta(p1,p2);
        if Length(connections)>2 then
        begin  {simulate a move of the middle connection point}
          connect1:=getconnection(connections[1]); {???}
          len:=dist(p1,connect.loc);
          newloc.x:=p1.x+len*cos(theta);
          newloc.y:=p1.y+len*sin(theta);
          connect1.loc:=newloc ;
        end;
      end
      else
      {we moved a middle point, treat it as a rotation about p1 and relocate p2}
      begin
        L.p1:=p1; L.p2:=newloc;
        extendline(L,dist(loc,p2));
        p2:=l.p2;
        Len:=dist(p1,p2);
        theta:=gettheta(p1,p2);
        connect1:=getconnection(connections[high(connections)]); {???}
        connect1.loc:=p2 ;
      end;
    end;


    {now move all the rods to the current connection points}
    for i:=0 to high(rods) do
    with rods[i] do
    begin
      connect1:=getconnection(connections[0]);
      p1:=connect1.loc;
      if connect1.movementtype = trace then connect1.resettrace;
      connect1:=getconnection(connections[high(connections)]);
      p2:=connect1.loc;
      if connect1.movementtype= trace then connect1.resettrace;
    end;
  end;
end;

{************ Tconnection.ResetTrace **********}
procedure TConnection.resetTrace;
begin
  setlength(tracepoints,0);
  tracecomplete:=false;
end;

{************* TConnection.setloc *************}
procedure TConnection.setloc(newloc:TRealpoint);
begin
  loc:=newloc;
  if (movementtype=trace) and (not tracecomplete) then
  begin
    prevpt:=pt;
    pt:=pointfromrealpoint(loc);
    if (Length(tracepoints)>0) then
    begin
      if (Length(tracepoints)>=10) and
         (prevpt.x=tracepoints[0].x) and (prevpt.y=tracepoints[0].y)
         and (pt.x=tracepoints[1].x) and (pt.y=tracepoints[1].y)

      then tracecomplete:=true  {a complete trace cycle has been made}
      else
      begin
        setlength(tracepoints,Length(tracepoints)+1);
        tracepoints[high(tracepoints)]:=pt;
      end;
    end
    else
    begin
      setlength(tracepoints,1);
      tracepoints[0]:=pt;
    end;
  end;
end;  

{************ TConnection.save *************}
procedure TConnection.save(fstream:TFilestream);
begin
  with fstream do
  begin
    writestr(fstream,ID);
    writebuffer(Loc, sizeof(loc));
    writebuffer(movementType, sizeof(movementtype));
    writebuffer(ConnectColor, sizeof(connectcolor));
  end;
end;

{********** TConnection.load ********}
procedure TConnection.load(fstream:TFilestream);
begin
  with fstream do
  begin
    id:=readstr(fstream);
    readbuffer(Loc, sizeof(loc));
    readbuffer(movementType, sizeof(movementtype));
    readbuffer(ConnectColor, sizeof(connectcolor));
  end;
end;


{*********** TConnections.Create ********}
constructor  TConnections.create;
begin
  inherited;
  sorted:=true;
end;


{********** TConnections.Save *********}
procedure TConnections.save(fstream:TFileStream);
var
  i:integer;
begin
  with fstream do
  begin
    i:=count;
    writebuffer(i,sizeof(i));
    for i:=0 to count-1 do
    begin
      TConnection(objects[i]).save(fstream);
    end;
  end;
end;

{********** TConnections.Load *********}
procedure TConnections.load(fstream:TFilestream);
var
  i,n:integer;
  connect:TConnection;
begin
  for i:=0 to count-1 do TConnection(Objects[i]).free;
  clear;
  with fstream do
  begin
    readbuffer(n, sizeof(n));
    for i:=0 to n-1 do
    begin
      connect:=TConnection.create;
      connect.load(fstream);
      addobject(connect.id, connect);
    end;
  end;
end;

destructor TRod.destroy;
begin
  force.free;
  inherited;
end;


{********** TRod.Save *********}
procedure TRod.save(fstream:TFileStream);
var
  i:integer;
begin
  with fstream do
  begin
    writestr(fstream,id);
    writebuffer(p1, sizeof(p1));
    writebuffer(p2, sizeof(p2));
    writebuffer(theta, sizeof(theta));
    writebuffer(Len, sizeof(len));
    writebuffer(rodcolor, sizeof(rodcolor));
    i:=length(connections);
    writebuffer(i, sizeof(i));
    for i:=0 to high(connections) do writestr(fstream,connections[i]);
    with force do
    begin
      i:=ord(FType);
      writebuffer(i,sizeof(i));
      writebuffer(Strength,sizeof(strength));
      writebuffer(dir,sizeof(dir));
    end;
    writebuffer(w, sizeof(w));
  end;
end;


procedure TRod.load(fstream:TFilestream);
var
  i:integer;
  nbrconnections:integer;
begin
  with fstream do
  begin
    id:=readstr(fstream);
    readbuffer(p1, sizeof(p1));
    readbuffer(p2, sizeof(p2));
    readbuffer(theta, sizeof(theta));
    readbuffer(Len, sizeof(len));
    readbuffer(rodcolor, sizeof(rodcolor));
    readbuffer(nbrconnections, sizeof(i));
    setlength(connections,nbrconnections);
    for i:=0 to high(connections) do connections[i]:=readstr(fstream);
    force:=TForce.create;
    with force do
    begin
      readbuffer(i,sizeof(i));
      FType:=TForcetype(i);
      readbuffer(Strength,sizeof(strength));
      readbuffer(dir,sizeof(dir));
    end;
    readbuffer(w, sizeof(w));
  end;
end;


{******** TLinkage.FindFixedConnection **********}
function TLinkage.FindFixedConnection(rod:TRod):integer;
{Find the fixed connection for a rod (if it exists)}
var
  i:integer;
begin
  with rod do
  begin
    result:=-1;
    for i:=0 to high(connections) do
    begin
      with getconnection(connections[i]) do
      begin
        if movementtype=pinned then
        begin
          result:=i;
          break;
        end;
      end;
    end;
  end;
end;



{********* TLinkage.DrawLinks *********}
procedure TLinkage.DrawLinks;
var
  i,j,n :integer;
  connect:TConnection;
  ip1,ip2:TPoint;
  L,L1,L2:TLine;
  Pid2:extended;
  wd2:integer;
  dx,dy:integer;
begin
  for n:= 0 to high(rods) do
  with rods[n], {target,}canvas do
  begin
    pid2:=pi/2;
    wd2:=w div 2;
    pen.width:=2;
    pen.Color:=rodcolor;
    {make integer points for plotting}
    ip1:= point(round(p1.x),round(p1.Y));
    ip2:= point(round(p2.x),round(p2.Y));
    L:=line(ip1,ip2);
    {find the perpemdicular line at p1 end}
    L1:=angledlinefromline(L,ip1,wd2, pid2); {from center line to one side}
    {get offsets for ends of this line}
    dx:=L1.p1.x-L1.p2.x-1;
    dy:=L1.p1.y-L1.p2.y-1;

    L1.p1.x:=L1.p2.X+2*dx+1; {apply 2*offsets to get corner point for other side}
    L1.p1.Y:=L1.p2.Y+2*dy+1;
    {other end of rod can use same offsets t define end line}
    L2.p2.X:=L.p2.x-dx;
    l2.p2.Y:=L.p2.Y-dy;
    L2.p1.X:=L2.p2.x+2*dx+1;
    l2.p1.Y:=L2.p2.Y+2*dy+1;

    {Now draw the rod}
    with L1 do moveto(p1.x,p1.y);
    with L1 do lineto(p2.x,p2.y);
    with l2 do lineto(p2.x,p2.y);
    with l2 do lineto(p1.x,p1.y);
    with L1 do lineto(p1.x,p1.y);

    {Draw the connections}
    pen.width:=1;
    for i:= 0 to high(connections) do
    begin
      connect:=getconnection(connections[i]);
      if connect<>nil then
      with connect do
      begin
        if movementType=pinned then
        begin
          brush.color:=connectcolor {clmaroon};
          with loc do rectangle(round(X)-5,round(y)-5,round(x)+5,round(y)+5);
        end
        else
        begin
          brush.color:=connectcolor {cllime};
          with loc do ellipse(round(X)-4,round(y)-4,round(x)+4,round(y)+4);
        end;
        if (movementtype=Trace) and (Length(tracepoints)>0) then
        begin
          with tracepoints[0] do canvas.moveto(x,y);
          for j := 1 to high(tracepoints) do
          with tracepoints[j] do lineto(x,y);
        end;
      end;
    end;
  end;
end;

{************** TLinkage.MouseDown *************}
procedure TLinkage.LinkMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);

var
  i,j,k:integer;
  connect:TConnection;
begin
  {see if mouse is on or near a defined connection point]
  {If so identify the point}
  {If not, but near a line, ask if connection point is to be added}
    {check for click on a connection point}
  if (ssleft in shift) or (ssright in shift) then
  begin
    with connectlist do
    if count>0 then
    for i:=0 to count-1 do
    begin
      connect:=TConnection(objects[i]);
      with connect do
      if ssleft in shift then
      begin
        if (abs(loc.X-x)<=5) and (abs(loc.Y-y)<=5) then
        begin
          dragmode:=modify;
          dragged:=connect;
          setlength(draggedrods,0);
          for j:=0 to high(rods) do
          with rods[j] do
          for k:=0 to high(connections) do
          begin
            if connections[k]=connect.id then
            begin
              setlength(draggedrods,Length(draggedrods)+1);
              with draggedrods[high(draggedrods)] do
              begin
                nbr:=j;
                if (p1.x=loc.x) and (p1.y=Loc.Y) then Pnbr:=1
                else if (p2.x=loc.x) and (p2.y=Loc.Y) then Pnbr:=2
                else Pnbr:=0;
              end;
            end;
          end;
          break;
        end;
      end
      else {right button depressed}
      if (abs(loc.X-x)<=5) and (abs(loc.Y-y)<=5) then
      with connectdlg do
      begin
        dlgcolor.color:=connectcolor;
        if (movementtype<>pinned) then
        begin
          dlgtypebox.checked:=movementtype=trace;
          dlgtypebox.visible:=true;
        end
        else dlgtypebox.visible:=false;

        if connectdlg.showmodal = mrOK then
        begin
          connectcolor:=Dlgcolor.color;
          if (movementtype<>pinned) then
          begin
            if dlgtypebox.checked then movementtype:=trace
            else movementtype:=float;
            resettrace;
          end;
        end;
      end;
    end;
  end;
end;


{**************** TLinkage.MouseMove *************8}
procedure TLinkage.LinkMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  newpoint:TRealpoint;
begin
  if (ssleft in shift) and (dragged<>nil) then
  with dragged do
  begin
    newpoint.x:=x;
    newpoint.y:=y;
    moveto(dragged,newpoint);
  end;
  invalidate;
end;

{*************** TLinkage.MouseUp **************8}
procedure TLinkage.LinkMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  dragged:=nil;
  dragmode:=nodrag;
  setlength(draggedrods,0);
end;

{**************** TLinkage.Save ************8}
procedure TLinkage.Save(fname:string);
{Save the current linkage definition}
var
  i,j:integer;
  rodcount:integer;
  fstream:TFileStream;
begin
  try
    fstream:=TFileStream.create(fname,fmCreate);
    with fstream do
    begin
      writestr(fstream,'V1');  {Verison 1 of saved linkage format}
      connectList.save(fstream);
      rodcount:=length(rods);
      writebuffer(rodcount,sizeof(rodcount));
      for i:=0 to high(rods) do rods[i].save(fstream);
    end;
  finally
    fstream.free;
  end;
end;

{**************** TLinkage.Load ************8}
procedure TLinkage.Load(fname:string);
{Load a linkage definition}
var
  i,j:integer;
  rodcount:integer;
  fstream:TFileStream;
  version:string;
begin
  try
    fstream:=TFileStream.create(fname,fmOpenRead);
    with fstream do
    begin
      version:=readstr(fstream);  {Verison 1 of saved linkage format}
      if version='V1' then
      begin
        connectList.load(fstream);
        for i:=0 to high(rods) do rods[i].free;
        readbuffer(rodcount,sizeof(rodcount));
        setlength(rods,rodcount);
        for i:=0 to rodcount-1 do
        begin
          rods[i]:=TRod.create;
          rods[i].load(fstream);
        end;
      end;
    end;
  finally
    fstream.free;
  end;
end;


{********** TLinkage.MakeConnection *******8}
function TLinkage.makeconnection(newid:string;newloc:TRealPoint;
                             newmovement:TMovementType; newcolor:TColor):integer;
{Create  a new connection}
var
  connect:TConnection;
begin
  connect:=TConnection.create;
  with connect do
  begin
    id:=newid;
    loc:=newloc;
    movementtype:=newmovement;
    connectColor:=newcolor;
    resettrace;
  end;
  result:=connectlist.AddObject(newid,connect);
end;

constructor TLinkage.create(Aowner:TComponent; Proto:TPaintbox);
begin
  inherited create(Aowner);
  connectlist:=TConnections.create;
  onmousedown:=LinkMousedown;
  OnMousemove:=LinkMousemove;
  OnMouseup:=LinkMouseup;
  OnPaint:=Paint;
  height:=proto.height;
  width:=proto.width;
  left:=proto.left;
  top:=proto.top;
  parent:=proto.parent;
end;

{************** TLinkage.GetConnection **********}
function TLinkage.getconnection(s:string):TConnection;
var
  index:integer;
begin
  if connectlist.find(s,index)
  then  result:=TConnection(connectlist.objects[index])
  else result:=nil;
end;

{********** TLinkage.destroy *************}
destructor TLinkage.destroy;
var
  i,j,index:integer;
begin
  for j:=0 to high(rods) do
  with rods[j] do
  for i:=0 to high(connections) do
  begin
    if connectlist.Find(connections[i],index) then
    begin
      TConnections(connectlist.Objects[index]).free;
      connectlist.delete(index);
    end;
  end;
  inherited;
end;

{*********** TLinkage.Paint *********}
procedure TLinkage.Paint(Sender: TObject);
var
  i:integer;
begin
  with canvas do
  begin
    brush.color:=clwhite;
    rectangle(clientrect);
    DrawLinks;
  end;
end;


end.
