unit U_FourBarLinkage3;
{Copyright  2008, Gary Darby,  www.DelphiForFun.org
 This program may be used or modified for any non-commercial purpose
 so long as this original notice remains in place.
 All other rights are reserved
 }


{Version 1 - a basic 4 bar "windshield wiper" linkage.}
{Version 2 adds user mouse operations to change a linkage diagram}
{Version 3 - define additional linkage points,
             "stretch, stop,  or disconnect" option when bind points are
                      encountered,
             add momentum to allow "double crank operations
             add trace point option
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  shellAPI, StdCtrls, ExtCtrls, UGeometry, ComCtrls, Menus, ULinkage;

type

  TForm1 = class(TForm)
    StaticText1: TStaticText;
    BuildTestBtn: TButton;
    PaintBox1: TPaintBox;
    RunBtn: TButton;
    Speedbar: TTrackBar;
    Memo1: TMemo;
    Savebtn: TButton;
    LoadBtn: TButton;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    Label1: TLabel;

    procedure FormActivate(Sender: TObject);
    procedure BuildTestBtnClick(Sender: TObject);
    procedure RunBtnClick(Sender: TObject);
    procedure SavebtnClick(Sender: TObject);
    procedure LoadBtnClick(Sender: TObject);
    procedure StaticText1Click(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
  public
    filename:string;
    Linkage:TLinkage;
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

var
  TwoPi:extended=2*Pi;

{************* FormActivate ***********}
procedure TForm1.FormActivate(Sender: TObject);
begin
  doublebuffered:=true;
  opendialog1.initialdir:=extractfilepath(application.exename);
  savedialog1.initialdir:=opendialog1.initialdir;
  buildtestbtnclick(sender);
end;


{************ BuildTestBtn ***********}
procedure TForm1.BuildTestBtnClick(Sender: TObject);
var
  i,index:integer;
  connect:TConnection;
  s:string;
begin
  linkage.free;
  linkage:=TLinkage.create(self,paintbox1);
  {make 3 rods, the implied 4th is between the two fixed points}
  with linkage do
  begin
    setlength(rods,3);
    for i:=0 to high(rods) do
    begin
      rods[i]:=TRod.create;
      with rods[i] do
      begin
        Force:=TForce.create;
        force.FType:=none;
        w:=8;
        rodcolor:=clblack;
      end;
    end;
    with connectlist, rods[0] do
    {the "crank" driving rod}
    begin
      id:='R0';
      rodcolor:=clred;
      p1:=realpoint(100,350);
      p2:=realpoint(100,315);
      theta:=gettheta(p1,p2);
      Len:=dist(p1,p2);
      setlength(Connections,2);
      s:='Sw1';
      connections[0]:=s;
      if not connectlist.find(s,index)
      then makeconnection(s,p1,pinned, clmaroon);
      s:='Sw2';
      connections[1]:=s;
      if not connectlist.find(s,index)
      then makeconnection(s,p2,trace,cllime);
      with force do
      begin
        FType:=rotational;
        dir:=+1;
        strength:=1;
      end;
    end;

    {Make connecting rod}
    with connectlist, rods[1]do
    begin
      id:='R1';
      rodcolor:=clBlue;
      connect:=getconnection('Sw2');
      p1:=connect.loc;
      p2:=realpoint(250,300);
      theta:=gettheta(p1,p2);
      Len:=dist(p1,p2);
      setlength(Connections,2);
      connections[0]:='Sw2';
      s:='Sw3';
      connections[1]:=s;
      if not connectlist.find(s,index)
      then  makeconnection(s,p2,float, clblue);
    end;

    {make the 3rd rod}
    with connectlist, rods[2]{,line} do
    begin
      id:='R2';
      Rodcolor:=clgreen;
      p2:=realpoint(250,200);
      p1:=realpoint(250,350);
      theta:=gettheta(p1,p2);
      Len:=dist(p1,p2);
      setlength(Connections,3);
      connections[1]:='Sw3';
      s:='Sw4';
      connections[0]:=s;
      if not connectlist.find(s,index)
      then  makeconnection(s,p1,pinned,clmaroon); {the fixed pivot point}
      s:='Sw5';
      connections[2]:=s;
      if not connectlist.find(s,index)
      then  makeconnection(s,p2,float, clFuchsia); {the upper pivot point}
    end;
    (*  {make the truss version}
    setlength(rods,5);
    for i:=3 to high(rods) do
    begin
      rods[i]:=TRod.create;
      with rods[i] do
      begin
        Force:=TForce.create;
        force.FType:=none;
        w:=8;
        rodcolor:=clblack;
      end;
    end;
    {Make 4th rod} {part of triangle forme by rods 2,4, & 5}
    with connectlist, rods[3]do
    begin
      id:='R3';
      Rodcolor:=clgreen;
      p2:=realpoint(175,250);
      connect:=getconnection('Sw2');
      p1:=connect.loc;
      theta:=gettheta(p1,p2);
      Len:=dist(p1,p2);
      setlength(Connections,2);
      connections[0]:='Sw2';
      connections[1]:='Sw6';
      makeconnection('Sw6',p2,trace,clSkyblue);
    end;

    {Make 5th rod} {part of triangle forme by rods 2,4, & 5}
    with connectlist, rods[4]do
    begin
      id:='R4';
      Rodcolor:=clskyblue;
      connect:=getconnection('Sw3');
      p1:=connect.loc;
      connect:=getconnection('Sw6');
      p2:=connect.loc;
      theta:=gettheta(p1,p2);
      Len:=dist(p1,p2);
      setlength(Connections,2);
      connections[0]:='Sw3';
      connections[1]:='Sw6';
    end;
    *)
    invalidate;
  end;
end;




function sign(n:integer):integer;
begin  if n<0 then result:=-1 else result:=+1;  end;

{************* RunBtn **********}
procedure TForm1.RunBtnClick(Sender: TObject);
var
   L:TRealLine;
   connect1,connect2:TConnection;
   circle1,circle2:TRealCircle;
   p:Trealpoint;
   ip1,ip2:TrealPoint;
   extendlen:extended;
   i,n:integer;
begin
  tag:=(tag+1) mod 2; {Toggle "tag" property on each entry}
  if tag=1 then
  begin
    Runbtn.caption:='Stop';
    with linkage do
    repeat
      for i:=0 to connectlist.count-1
      do TConnection(connectlist.objects[i]).moved:=false;
      n:=0;
      {move rod 0 - the "driver" rod}
      with rods[n] do
      begin
        if force.ftype=rotational then
        begin
          {rotate the "driver" 1 degree for each strength unit }
          theta:=theta+degtorad(force.strength*sign(form1.speedbar.position));
          if theta>TwoPi then theta:=theta-Twopi;
          rotaterightendto(p1,p2,theta); {rotate end p2 about p1 by theta}
          {update the hinge point}
          connect1:=getconnection(connections[high(connections)]);
          connect1.setloc(p2);
          connect1.moved:=true;
          inc(n);
          if n<=high(rods) then
          {move rod 1 - the "connecting" rod}
          with rods[n] do
          begin
            connect1:=getconnection(connections[0]);
            p1:=connect1.loc; {the right end hinge point}
            circle1:=circle(p1.x,p1.y,Len); {other end must lie on this circle}
            connect2:=getconnection(connections[high(connections)]);
            p:=rods[N+1].p1;
            {it also must lie on the circle that would be drawn by the windhsield
             wiper rod  at the left hand end point}
            circle2:=circle(rods[n+1].p1.x,rods[N+1].p1.y,
                                        dist(p,connect2.loc));
            if circlecircleIntersect(circle1, circle2,IP1,Ip2)
            {IP1 is where the circles intersect, hopefully - might
             fall on Ip2 under some configurations?}
            then
            begin
              connect2.setloc(ip1);
              p2:=ip1;
            end;
            connect2.moved:=true;
          end;
          inc(n);
          if n<=high(rods) then
          {move rod 2 - the "wiper" rod}
          with rods[n]do
          begin
            connect1:=getconnection(connections[0]);
            p1:=connect1.loc; {the right end hinge point}
            circle1:=circle(p1.x,p1.y,Len); {other end must lie on this circle}
            if high(connections)>1 then
            begin
              connect2:=getconnection(connections[1]);
              extendLen:=round(Len-dist(p1,connect2.Loc));
              L:=line(p1,connect2.loc);
              extendline(L,Extendlen);
              p2:=L.p2;
              connect1:=getconnection(connections[high(connections)]);
              connect1.setloc(p2);
              connect1.moved:=true;
            end
            else   showmessage('???');
          end;
          inc(n);
          if n<=high(rods) then
          {move rods 4 & 5 }
          with rods[n]do
          begin
            connect1:=getconnection(connections[0]);
            p1:=connect1.loc; {the right end hinge point}
            circle1:=circle(p1.x,p1.y,Len); {other end must lie on this circle}

            if high(connections)=1 then
            with rods[n+1] do
            begin
              connect1:=getconnection(connections[0]);
              p1:=connect1.loc; {the right end hinge point}
              circle2:=circle(p1.x,p1.y,Len); {other end must lie on this circle}
              if circlecircleIntersect(circle1, circle2,IP1,Ip2)
              {IP1 is where the circles intersect, hopefully - might
               fall on Ip2 under some configurations?}
              then
              begin
                connect2:=getconnection(connections[1]);
                connect2.setloc(ip1);
                p2:=ip1;
                with connect2 do
                begin
                  (* {tracing moved to setloc procedure}
                  if (movementtype=trace) and (not tracecomplete) then
                  begin
                    pt:=pointfromrealpoint(loc);
                    if (Length(tracepoints)>0) then
                    begin
                      if (Length(tracepoints)>=10) and
                         (pt.x=tracepoints[0].x) and (pt.y=tracepoints[0].y)

                      then tracecomplete:=true  {a complete trace cycle has been made}
                      else
                      begin
                        setlength(tracepoints,Length(tracepoints)+1);
                        tracepoints[high(tracepoints)]:=pt;
                      end;
                    end
                    else
                    begin
                      setlength(tracepoints,1);
                      tracepoints[0]:=pt;
                    end;
                  end;
                  *)
                end;
              end;
              connect2.moved:=true;

            end
            else   showmessage('???');
            p2:=ip1;
          end;
        end;
      end;
      sleep(100-abs(form1.speedbar.position));
      invalidate; {tell Windows that the image needs to be redrawn}
      application.processmessages; {allow handling redraw and Stop button click messages}
    until self.tag=0;
  end
  else runbtn.caption:='Run';
end;


{************** SavebtnClick **************}
procedure TForm1.SavebtnClick(Sender: TObject);
begin
  if savedialog1.execute then
  begin
    filename:=savedialog1.filename;
    linkage.Save(filename);
  end;
end;

{************** LoadbtnClick **********}
procedure TForm1.LoadBtnClick(Sender: TObject);
begin
  if opendialog1.execute then
  begin
    filename:=opendialog1.filename;
    linkage.Load(filename);
    linkage.invalidate;
  end;
end;

{*************** FormCloseQuery *********88}
procedure TForm1.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  tag:=0; {in case linkage is running, tag=0 stops it}
  application.processmessages;
end;

procedure TForm1.StaticText1Click(Sender: TObject);
begin
  ShellExecute(Handle, 'open', 'http://www.delphiforfun.org/',
  nil, nil, SW_SHOWNORMAL) ;
end;

end.
